﻿using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  public static class IntervalUtil {
    public static IEnumerable<double> GetConstraintViolations(
      IEnumerable<ShapeConstraint> constraints, IBoundsEstimator estimator, IntervalCollection intervalCollection,
      ISymbolicExpressionTree solution) {
      return constraints.Select(constraint => GetConstraintViolation(constraint, estimator, intervalCollection, solution)).ToList();
    }

    public static double GetConstraintViolation(
      ShapeConstraint constraint, IBoundsEstimator estimator, IntervalCollection variableRanges,
      ISymbolicExpressionTree tree) {
      var varRanges = variableRanges.GetReadonlyDictionary();

      if (constraint.Variable != null && !varRanges.ContainsKey(constraint.Variable)) {
        throw new ArgumentException(
          $"The given variable {constraint.Variable} in the constraint does not exist in the model.",
          nameof(constraint));
      }

      // Create new variable ranges for defined regions
      var regionRanges = new IntervalCollection();
      foreach (var kvp in varRanges) {
        if (constraint.Regions.GetReadonlyDictionary().TryGetValue(kvp.Key, out var val)) {
          regionRanges.AddInterval(kvp.Key, val);
        } else {
          regionRanges.AddInterval(kvp.Key, kvp.Value);
        }
      }

      if (!constraint.IsDerivative) {
        return estimator.GetConstraintViolation(tree, regionRanges, constraint);
      } else {
        for (var i = 0; i < constraint.NumberOfDerivations; ++i) {
          if (!estimator.IsCompatible(tree) || !DerivativeCalculator.IsCompatible(tree)) {
            throw new ArgumentException("The tree contains an unsupported symbol.");
          }

          tree = DerivativeCalculator.Derive(tree, constraint.Variable);
        }

        return estimator.GetConstraintViolation(tree, regionRanges, constraint);
      }
    }

  }
}